import json
import multiprocessing
import time
from abc import ABC
from datetime import datetime, timedelta
from enum import Enum
from functools import partial
from typing import Dict, Any, Tuple, List

import boto3
import pandas as pd
import pytz
import requests
from botocore.exceptions import ClientError
from loguru import logger as lg
import os


class SecretManager:

    _secretClient: boto3.client

    def __init__(self, region: str = 'us-east-1'):
        self._secretClient = boto3.client('secretsmanager', region_name=region)

    def get_secret(self, secretName: str):
        try:
            secretResponse = self._secretClient.get_secret_value(SecretId=secretName)
        except ClientError as e:
            # For a list of exceptions thrown, see
            # https://docs.aws.amazon.com/secretsmanager/latest/apireference/API_GetSecretValue.html
            raise e

        secretString = secretResponse['SecretString']
        return secretString

class ApiKeyGetMode(Enum):
    LOCAL_FILE = "local_file"
    AWS_SECRET_MANAGER = "aws_secret_manager"

class NoDataReturned(Exception):
    def __init__(self, message="No data returned from the API"):
        self.message = message
        super().__init__(self.message)

class RestService(ABC):
    _key_get_mode: ApiKeyGetMode
    _local_keys_path: str
    _aws_secret_name: str
    _aws_secret_key: str
    _max_threads: int

    def __init__(self, api_key_get_mode: ApiKeyGetMode, key_get_params: Dict, max_threads: int = 32):
        self._key_get_mode = api_key_get_mode
        if api_key_get_mode == ApiKeyGetMode.LOCAL_FILE:
            self._local_keys_path = key_get_params['local_key_path']
            self._aws_secret_name = None
            self._aws_secret_key = None
        elif api_key_get_mode == ApiKeyGetMode.AWS_SECRET_MANAGER:
            self._local_keys_path = None
            self._aws_secret_name = key_get_params['aws_secret_name']
            self._aws_secret_key = key_get_params['aws_secret_key']
        else:
            raise ValueError(f"Invalid API Key Get Mode: {api_key_get_mode}")
        self._max_threads = max_threads

    def _headers(self) -> Dict[str, str]:
        if self._key_get_mode == ApiKeyGetMode.LOCAL_FILE:
            amberdata_api_key = get_amberdata_api_key_from_local_file(self._local_keys_path)
        elif self._key_get_mode == ApiKeyGetMode.AWS_SECRET_MANAGER:
            amberdata_api_key = get_amberdata_api_key_from_aws_secret_manager(self._aws_secret_name, self._aws_secret_key)
        else:
            raise ValueError(f"Invalid API Key Get Mode: {self._key_get_mode}")
        return {'accept': 'application/json', 'x-api-key': f'{amberdata_api_key}'}

    def _get_max_threads(self) -> int:
        return self._max_threads

    @staticmethod
    def _get_date_ranges_for_parallel(start_date: datetime, end_date: datetime, batch_period: timedelta) -> List[Tuple[datetime, datetime]]:
        date_ranges = []
        batch_start_date = start_date
        batch_end_date = min(batch_start_date + batch_period, end_date)
        while batch_start_date < end_date:
            date_ranges.append((batch_start_date, batch_end_date))
            batch_start_date = batch_end_date
            batch_end_date = min(batch_start_date + batch_period, end_date)

        return date_ranges

    @staticmethod
    def _validate_response(response: dict, request_description: str, url: str, params: Dict) -> bool:
        if 'status' not in response:
            lg.error(f"Error on request: {request_description}, url: {url}, params: {params}")
            lg.error(f"No 'status' found in response: {response}")
        elif 'title' not in response:
            lg.error(f"Error on request: {request_description}, url: {url}, params: {params}")
            lg.error(f"No 'title' found in response: {response}")
        elif 'description' not in response:
            lg.error(f"Error on request: {request_description}, url: {url}, params: {params}")
            lg.error(f"No 'description' found in response: {response}")
        else:
            status = response['status']
            title = response['title']
            description = response['description']
            if status != 200 or title != "OK" or description != "Successful request":
                error = "N/A"
                if 'error' in response:
                    error = response['error']
                message = response['message']
                lg.error(f"Error on request: {request_description}, url: {url}, params: {params}")
                lg.error(
                    f"{request_description} failed.\nStatus:{status}, Title:{title},"
                    f" Desc:{description}, Error:{error}, Message:{message}")
                return False
            else:
                return True
        return False

    @classmethod
    def _get_response(cls, url: str, params: Dict[str, str], headers: Dict[str, str], description: str,
                      retry_count: int = 5, sleep_duration: float = 10.0) -> [bool, object]:
        while retry_count > 0:
            lg.debug(f"Executing request with url:{url}, params={params}, retryCount:{retry_count}")
            with requests.session() as session:
                try:
                    response_raw = session.get(url, headers=headers, params=params)
                    response_raw.raise_for_status()
                    response_data = json.loads(response_raw.text)
                    clean_response = cls._validate_response(response_data, description, url, params)
                    if not clean_response:
                        retry_count -= 1
                        time.sleep(sleep_duration)
                    else:
                        return clean_response, response_data
                except requests.exceptions.ContentDecodingError as e:
                    lg.error(f"Decoding error: {e}")
                    retry_count -= 1
                    time.sleep(sleep_duration)
                except requests.exceptions.RequestException as e:
                    lg.error(f"Request failed: {e}")
                    retry_count -= 1
                    time.sleep(sleep_duration)
        return False, None

    @classmethod
    def _process_response(cls, response: dict) -> Dict[str, Any]:
        # Process payload with metadata
        return_dict = {}
        if (type(response['payload']) is dict) and ('metadata' in response['payload'].keys()):
            return_dict['metadata'] = response['payload']['metadata']
            return_dict['data'] = response['payload']['data']
            return return_dict
        else:
            return_dict['data'] = response['payload']
        return return_dict

    @staticmethod
    def _process_payload_df(payload) -> pd.DataFrame:
        if 'metadata' in payload.keys():
            # We are processing a historical data payload, handle metadata
            # Separate the actual data first
            sub_payload = payload['data']
            # If no data, return empty data frame
            if len(sub_payload) == 0:
                return pd.DataFrame()
            elif 'columns' in payload['metadata']:
                columns = payload['metadata']['columns']
                _df = pd.DataFrame.from_dict(sub_payload)
                _df.columns = columns
            else:
                _df = pd.DataFrame.from_dict(sub_payload)

            for key in payload['metadata'].keys():
                # Columns would have been handled earlier
                if key == "columns":
                    continue
                _df[key] = payload['metadata'][key]
        else:
            if type(payload['data']) == dict:
                _df = pd.DataFrame.from_records(payload['data'], index=[0])
            else:
                _df = pd.DataFrame.from_dict(payload['data'])

        tz_local = pytz.timezone('US/Eastern')
        # Add UTC & EST Time Columns
        if not _df.empty and 'timestamp' in _df.columns:
            _df['timeUTC'] = pd.to_datetime(pd.to_numeric(_df["timestamp"]), unit="ms", utc=True)
            _df['timeEST'] = _df['timeUTC'].dt.tz_convert(tz_local)
        elif not _df.empty and 'exchangeTimestamp' in _df.columns:
            _df['timeUTC'] = pd.to_datetime(pd.to_numeric(_df["exchangeTimestamp"]), unit="ms", utc=True)
            _df['timeEST'] = _df['timeUTC'].dt.tz_convert(tz_local)
        return _df

    @staticmethod
    def _process_payload_dict(payload) -> Dict:
        return_dict = {}
        sub_payload = payload['data']
        if len(sub_payload) == 0:
            return return_dict
        elif 'metadata' in payload.keys():
            for key in payload['metadata']:
                return_dict[key] = payload['metadata'][key]
        return_dict['data'] = sub_payload
        return return_dict

    @staticmethod
    def _process_batch(date_tuple: Tuple[datetime, datetime],
                       headers: Dict[str, str],
                       url: str,
                       params: Dict,
                       description: str):
        params['startDate'] = date_tuple[0].isoformat(timespec='milliseconds')
        params['endDate'] = date_tuple[1].isoformat(timespec='milliseconds')
        lg.info(f"Starting request for Start:{params['startDate']} to End: {params['endDate']}")
        _df = RestService.get_and_process_response_df(url, params, headers, description)
        lg.info(f"Finished request for Start:{params['startDate']} to End: {params['endDate']}")
        if not _df.empty:
            _df.reset_index(inplace=True)
            # drop the index column
            if 'index' in _df.columns:
                _df.drop(columns=['index'], inplace=True)
        return _df

    @staticmethod
    def _process_parallel(start_date: datetime, end_date: datetime, batch_period: timedelta, headers: Dict[str, str],
                          url: str, params: Dict, description: str, max_threads: int):
        # Use 1 less CPU Than available for safety
        cpu_count = min(max_threads, multiprocessing.cpu_count())
        lg.debug(f"Will use {cpu_count} threads")
        date_ranges = RestService._get_date_ranges_for_parallel(start_date, end_date, batch_period)
        partial_process_batch = partial(RestService._process_batch,
                                      headers=headers, url=url,
                                      params=params, description=description)
        p = multiprocessing.Pool(cpu_count)
        lg.debug("Starting multi threaded requests...")
        result_dfs = p.map(partial_process_batch, date_ranges)
        lg.debug("Finished multi threaded requests...")
        result_df = pd.concat([pd.DataFrame()] + result_dfs, ignore_index=True)
        if result_df.empty:
            lg.warning("No data returned from any of the parallel requests.")
        return result_df

    @classmethod
    def get_and_process_response_df(cls, url: str, params: Dict[str, str], headers: Dict[str, str], description: str,
                                    retryCount: int = 5) -> pd.DataFrame:
        more_data_to_fetch = True
        next_url = url
        paged_data = pd.DataFrame()
        while more_data_to_fetch:
            success, res = cls._get_response(next_url, params, headers, description, retryCount)
            if success:
                raw_data = cls._process_payload_df(cls._process_response(res))
                # If 'next' is in the raw data, and it contains a URL, it implies the data is paginated, run in a loop until all data is received.
                if 'next' in raw_data.columns and raw_data['next'].unique()[0] is not None:
                    next_url = raw_data['next'].unique()[0]
                    lg.debug(f"Fetching next page from: {next_url}")
                else:
                    more_data_to_fetch = False
                if not raw_data.empty:
                    paged_data = pd.concat([paged_data, raw_data], ignore_index=True)
                else:
                    continue
            else:
                raise ValueError(f"Failed to fetch data for request:{description}, url:{url}, params:{params} after {retryCount} retries.")

        if 'next' in paged_data.columns:
            paged_data.drop(columns=['next'], inplace=True)
        return paged_data

    @classmethod
    def get_and_process_response_dict(cls, url: str, params: Dict[str, str], headers: Dict[str, str], description: str,
                                      retry_count: int = 5) -> Dict:
        more_data_to_fetch = True
        next_url = url
        paged_data = []
        while more_data_to_fetch:
            success, res = cls._get_response(next_url, params, headers, description, retry_count)
            if success:
                raw_data = cls._process_payload_dict(cls._process_response(res))
                # If 'next' is in the raw data, and it contains a URL, it implies the data is paginated, run in a loop until all data is received.
                if 'next' in raw_data.keys() and raw_data['next'] is not None:
                    next_url = raw_data['next']
                    lg.debug(f"Fetching next page from: {next_url}")
                else:
                    more_data_to_fetch = False
                if len(raw_data) > 0 and 'data' in raw_data.keys():
                    paged_data.append(raw_data['data'])
                else:
                    continue
            else:
                raise ValueError(f"Failed to fetch data for request:{description}, url:{url}, params:{params} after {retry_count} retries.")

        # paged_data contains paginated data, we need to merge them properly but:
        # each page could be a list, at which point we'd simply do an append OR
        # each page could be a dict, at which point we need to merge each key in the dict
        ret_data = {}
        for page_data in paged_data:
            if type(page_data) == list:
                if 'data' not in ret_data.keys():
                    ret_data['data'] = page_data
                else:
                    ret_data['data'] = ret_data['data'] + page_data
            elif type(page_data) == dict:
                if 'data' not in ret_data.keys():
                    ret_data['data'] = {}
                for key in page_data.keys():
                    if key not in ret_data['data'].keys():
                        ret_data['data'][key] = page_data[key]
                    else:
                        ret_data['data'][key].update(page_data[key])
            else:
                raise Exception("Not implemented")
        return ret_data




def get_amberdata_api_key_from_local_file(file_path: str) -> str:
    try:
        with open(file_path, 'r') as file:
            keys = json.load(file)
            return keys.get('amberdata_api_key')
    except FileNotFoundError:
        full_path = os.path.abspath(file_path)
        raise FileNotFoundError(f"The file at {full_path} is missing.")
    except KeyError:
        raise KeyError("The 'amberdata_api_key' key is missing in the file.")
    except json.JSONDecodeError:
        raise ValueError("The file is not valid JSON.")

def get_amberdata_api_key_from_aws_secret_manager(secret_name: str, secret_key: str) -> str:
    sm = SecretManager()
    secret_string = sm.get_secret(secret_name)
    secret_dict = eval(secret_string)
    return secret_dict[secret_key]


import datetime
import json
from typing import List, Dict

import pandas as pd
from loguru import logger as lg

from ad_pipeline.common import RestService, NoDataReturned, ApiKeyGetMode
from ad_pipeline.constants import \
    AMBERDATA_FUTURES_REST_FUNDING_RATES_ENDPOINT, AMBERDATA_FUTURES_REST_LIQUIDATIONS_ENDPOINT, \
    AMBERDATA_FUTURES_REST_LONG_SHORT_RATIO_ENDPOINT, \
    TimeInterval, MarketDataVenue, TimeFormat, AMBERDATA_FUTURES_REST_INSURANCE_FUNDS_ENDPOINT, \
    SortDirection, AMBERDATA_FUTURES_REST_BATCH_FUNDING_RATES_ENDPOINT, AMBERDATA_FUTURES_REST_OHLCV_ENDPOINT, \
    AMBERDATA_FUTURES_REST_BATCH_OHLCV_ENDPOINT, AMBERDATA_FUTURES_REST_OPEN_INTEREST_ENDPOINT, \
    AMBERDATA_FUTURES_REST_BATCH_OPEN_INTEREST_ENDPOINT, AMBERDATA_FUTURES_REST_ORDER_BOOK_SNAPSHOTS_ENDPOINT, \
    AMBERDATA_FUTURES_REST_ORDER_BOOK_EVENTS_ENDPOINT, AMBERDATA_FUTURES_REST_TICKERS_ENDPOINT, \
    AMBERDATA_FUTURES_REST_TRADES_ENDPOINT, BatchPeriod


class FuturesRestService(RestService):

    def __init__(self, api_key_get_mode: ApiKeyGetMode, api_key_get_params: Dict, max_threads: int = 32):
        RestService.__init__(self, api_key_get_mode, api_key_get_params, max_threads)

    def get_funding_information_raw(self, exchanges: List[MarketDataVenue] = None, include_inactive: bool = None,
                                    time_format: TimeFormat = None) -> pd.DataFrame:
        params = {}
        if exchanges is not None and len(exchanges) > 0:
            params['exchange'] = ",".join(
                [exchange.value for exchange in exchanges])
        if include_inactive is not None:
            params['includeInactive'] = str(include_inactive).lower()
        if time_format is not None:
            params['timeFormat'] = time_format.value
        url = AMBERDATA_FUTURES_REST_FUNDING_RATES_ENDPOINT + "information"
        description = "FUTURES Funding Information Request"
        lg.info(f"Starting {description}")
        return_df = RestService.get_and_process_response_df(url, params, self._headers(), description)

        lg.info(f"Finished {description}")
        return return_df

    def get_funding_information(self, exchanges: List[MarketDataVenue] = None, include_inactive: bool = None,
                                time_format: TimeFormat = None, index_keys: List[str] = None) -> pd.DataFrame:
        if index_keys is None:
            index_keys = ['exchange']
        return self.get_funding_information_raw(exchanges, include_inactive, time_format).set_index(index_keys)

    def get_funding_rates_raw(self, instrument: str, exchange: MarketDataVenue, start_date: datetime = None,
                              end_date: datetime = None, time_format: TimeFormat = None,
                              batch_period: datetime.timedelta = BatchPeriod.HOUR_1.value,
                              parallel_execution: bool = False) -> pd.DataFrame:
        params = {'exchange': exchange.value}
        if start_date is not None:
            params['startDate'] = start_date.isoformat(timespec='milliseconds')
        if end_date is not None:
            params['endDate'] = end_date.isoformat(timespec='milliseconds')
        if time_format is not None:
            params['timeFormat'] = time_format.value
        url = AMBERDATA_FUTURES_REST_FUNDING_RATES_ENDPOINT + f"{instrument}"
        description = "FUTURES Funding Rates Request"
        lg.info(f"Starting {description}")
        if parallel_execution:
            if start_date is None or end_date is None:
                raise ValueError("Start and end date must be provided for parallel execution!")
            else:
                return_df = RestService._process_parallel(start_date, end_date, batch_period, self._headers(), url,
                                                          params, description, self._get_max_threads())
                # Check if timestamp or exchangeTimestamp + exchangeTimestampNano is present and sort by it otherwise skip sorting
                if 'timestamp' in return_df.columns:
                    return_df.sort_values('timestamp', inplace=True)
                elif 'exchangeTimestamp' in return_df.columns and 'exchangeTimestampNano' in return_df.columns:
                    return_df.sort_values(['exchangeTimestamp', 'exchangeTimestampNano'], inplace=True)
        else:
            return_df = RestService.get_and_process_response_df(url, params, self._headers(),
                                                                description)
        lg.info(f"Finished {description}")
        return return_df

    def get_funding_rates(self, instrument: str, exchange: MarketDataVenue, start_date: datetime = None,
                          end_date: datetime = None, time_format: TimeFormat = None,
                          batch_period: datetime.timedelta = BatchPeriod.HOUR_12.value, index_keys: List[str] = None,
                          parallel_execution: bool = False) -> pd.DataFrame:
        if index_keys is None:
            index_keys = ['exchange']
        return self.get_funding_rates_raw(instrument, exchange, start_date, end_date, time_format, batch_period, parallel_execution).set_index(index_keys)

    def get_funding_batch_historical_raw(self, exchange: MarketDataVenue, instruments: List[str],
                                         start_date: datetime = None,
                                         end_date: datetime = None, time_interval: TimeInterval = None,
                                         time_format: TimeFormat = None):
        params = {'exchange': exchange.value, 'instrument': ",".join(instrument for instrument in instruments)}
        if start_date is not None:
            params['startDate'] = start_date.isoformat(timespec='milliseconds')
        if end_date is not None:
            params['endDate'] = end_date.isoformat(timespec='milliseconds')
        if time_interval is not None:
            params['timeInterval'] = time_interval.value
        if time_format is not None:
            params['timeFormat'] = time_format.value
        url = AMBERDATA_FUTURES_REST_BATCH_FUNDING_RATES_ENDPOINT + f"exchange/{exchange.value}/" + "historical"
        description = "FUTURES Funding Rates Batch Historical Request"
        lg.info(f"Starting {description}")
        return_df = RestService.get_and_process_response_df(url, params, self._headers(),
                                                            description)
        lg.info(f"Finished {description}")
        return return_df

    def get_funding_batch_historical(self, exchange: MarketDataVenue, instruments: List[str],
                                     start_date: datetime = None,
                                     end_date: datetime = None, time_interval: TimeInterval = None,
                                     time_format: TimeFormat = None,
                                     index_keys: List[str] = None) -> pd.DataFrame:
        if index_keys is None:
            index_keys = ['instrument']
        return self.get_funding_batch_historical_raw(exchange, instruments, start_date, end_date, time_interval,
                                                     time_format).set_index(index_keys)

    def get_insurance_funds_information_raw(self, exchanges: List[MarketDataVenue] = None,
                                            include_inactive: bool = None,
                                            time_format: TimeFormat = None) -> pd.DataFrame:
        params = {}
        if exchanges is not None and len(exchanges) > 0:
            params['exchange'] = ",".join(
                [exchange.value for exchange in exchanges])
        if include_inactive is not None:
            params['includeInactive'] = str(include_inactive).lower()
        if time_format is not None:
            params['timeFormat'] = time_format.value
        url = AMBERDATA_FUTURES_REST_INSURANCE_FUNDS_ENDPOINT + "information"
        description = "FUTURES Insurance Funds Information Request"
        lg.info(f"Starting {description}")
        return_df = RestService.get_and_process_response_df(url, params, self._headers(), description)

        lg.info(f"Finished {description}")
        return return_df

    def get_insurance_funds_information(self, exchanges: List[MarketDataVenue] = None, include_inactive: bool = None,
                                        time_format: TimeFormat = None, index_keys: List[str] = None) -> pd.DataFrame:
        if index_keys is None:
            index_keys = ['exchange']
        return self.get_insurance_funds_information_raw(exchanges, include_inactive, time_format).set_index(index_keys)

    def get_insurance_funds_raw(self, instrument: str, exchange: MarketDataVenue, start_date: datetime = None,
                                end_date: datetime = None, time_format: TimeFormat = None,
                                sort_direction: SortDirection = None):
        params = {'exchange': exchange.value}
        if start_date is not None:
            params['startDate'] = start_date.isoformat(timespec='milliseconds')
        if end_date is not None:
            params['endDate'] = end_date.isoformat(timespec='milliseconds')
        if time_format is not None:
            params['timeFormat'] = time_format.value
        if sort_direction is not None:
            params['sortDirection'] = sort_direction.value
        url = AMBERDATA_FUTURES_REST_INSURANCE_FUNDS_ENDPOINT + f"{instrument}"
        description = "FUTURES Insurance Funds Request"
        lg.info(f"Starting {description}")
        return_df = RestService.get_and_process_response_df(url, params, self._headers(),
                                                            description)
        lg.info(f"Finished {description}")
        return return_df

    def get_insurance_funds(self, instrument: str, exchange: MarketDataVenue, start_date: datetime = None,
                            end_date: datetime = None, time_format: TimeFormat = None,
                            sort_direction: SortDirection = None, index_keys: List[str] = None) -> pd.DataFrame:
        if index_keys is None:
            index_keys = ['exchange']
        return self.get_insurance_funds_raw(instrument, exchange, start_date, end_date, time_format,
                                            sort_direction).set_index(index_keys)

    def get_liquidations_information_raw(self, exchanges: List[MarketDataVenue] = None, include_inactive: bool = None,
                                         time_format: TimeFormat = None) -> pd.DataFrame:
        params = {}
        if exchanges is not None and len(exchanges) > 0:
            params['exchange'] = ",".join(
                [exchange.value for exchange in exchanges])
        if include_inactive is not None:
            params['includeInactive'] = str(include_inactive).lower()
        if time_format is not None:
            params['timeFormat'] = time_format.value
        url = AMBERDATA_FUTURES_REST_LIQUIDATIONS_ENDPOINT + "information"
        description = "FUTURES Liquidations Information"
        lg.info(f"Starting {description}")
        return_df = RestService.get_and_process_response_df(url, params, self._headers(),
                                                            description)
        lg.info(f"Finished {description}")
        return return_df

    def get_liquidations_information(self, exchanges: List[MarketDataVenue] = None, include_inactive: bool = None,
                                     time_format: TimeFormat = None, index_keys: List[str] = None) -> pd.DataFrame:
        if index_keys is None:
            index_keys = ['exchange']
        return self.get_liquidations_information_raw(exchanges, include_inactive, time_format).set_index(index_keys)

    def get_liquidations_raw(self, instrument: str, exchange: MarketDataVenue, start_date: datetime = None,
                             end_date: datetime = None, time_format: TimeFormat = None,
                             sort_direction: SortDirection = None,
                             batch_period: datetime.timedelta = BatchPeriod.HOUR_12.value,
                             parallel_execution: bool = False) -> pd.DataFrame:
        params = {'exchange': exchange.value}
        if start_date is not None:
            params['startDate'] = start_date.isoformat(timespec='milliseconds')
        if end_date is not None:
            params['endDate'] = end_date.isoformat(timespec='milliseconds')
        if time_format is not None:
            params['timeFormat'] = time_format.value
        if sort_direction is not None:
            params['sortDirection'] = sort_direction.value
        url = AMBERDATA_FUTURES_REST_LIQUIDATIONS_ENDPOINT + f"{instrument}"
        description = "FUTURES Liquidations Request"
        lg.info(f"Starting {description}")

        if parallel_execution:
            if start_date is None or end_date is None:
                raise ValueError("Start and end date must be provided for parallel execution!")
            else:
                return_df = RestService._process_parallel(start_date, end_date, batch_period, self._headers(), url,
                                                          params, description, self._get_max_threads())
                # Check if timestamp or exchangeTimestamp + exchangeTimestampNano is present and sort by it otherwise skip sorting
                if 'timestamp' in return_df.columns:
                    return_df.sort_values('timestamp', inplace=True)
                elif 'exchangeTimestamp' in return_df.columns and 'exchangeTimestampNano' in return_df.columns:
                    return_df.sort_values(['exchangeTimestamp', 'exchangeTimestampNano'], inplace=True)
        else:
            return_df = RestService.get_and_process_response_df(url, params, self._headers(),
                                                                description)
        lg.info(f"Finished {description}")
        return return_df

    def get_liquidations(self, instrument: str, exchange: MarketDataVenue, start_date: datetime = None,
                         end_date: datetime = None, time_format: TimeFormat = None,
                         sort_direction: SortDirection = None, index_keys: List[str] = None,
                         batch_period: datetime.timedelta = BatchPeriod.HOUR_12.value,
                         parallel_execution: bool = False) -> pd.DataFrame:
        if index_keys is None:
            index_keys = ['exchange', 'instrument']
        return self.get_liquidations_raw(instrument,
                                         exchange, start_date, end_date, time_format, sort_direction,
                                         batch_period, parallel_execution).set_index(index_keys)

    def get_long_short_ratio_information_raw(self, exchanges: [MarketDataVenue] = None, include_inactive: bool = None,
                                             time_format: TimeFormat = None) -> pd.DataFrame:
        params = {}
        if exchanges is not None and len(exchanges) > 0:
            params['exchange'] = ",".join(
                [exchange.name for exchange in exchanges])
        if include_inactive is not None:
            params['includeInactive'] = str(include_inactive).lower()
        if time_format is not None:
            params['timeFormat'] = time_format.value
        url = AMBERDATA_FUTURES_REST_LONG_SHORT_RATIO_ENDPOINT + "information"
        description = "FUTURES Long Short Information Request"
        lg.info(f"Starting {description}")
        return_df = RestService.get_and_process_response_df(url, params, self._headers(),
                                                            description)
        lg.info(f"Finished {description}")
        return return_df

    def get_long_short_ratio_information(self, exchanges: [MarketDataVenue] = None, include_inactive: bool = None,
                                         time_format: TimeFormat = None, index_keys: List[str] = None) -> pd.DataFrame:
        if index_keys is None:
            index_keys = ['exchange']
        return self.get_long_short_ratio_information_raw(exchanges, include_inactive, time_format).set_index(index_keys)

    def get_long_short_ratio_raw(self, instrument: str, exchange: MarketDataVenue = None, start_date: datetime = None,
                                 end_date: datetime = None, time_format: TimeFormat = None, time_interval: TimeInterval = None,
                                 sort_direction: SortDirection = None, batch_period: datetime.timedelta = BatchPeriod.HOUR_1.value,
                                 parallel_execution: bool = False) -> pd.DataFrame:
        params = {'exchange': exchange.value}
        if start_date is not None:
            params['startDate'] = start_date.isoformat(timespec='milliseconds')
        if end_date is not None:
            params['endDate'] = end_date.isoformat(timespec='milliseconds')
        if time_format is not None:
            params['timeFormat'] = time_format.value
        if time_interval is not None:
            params['timeInterval'] = time_interval.value
        if sort_direction is not None:
            params['sortDirection'] = sort_direction.value

        url = AMBERDATA_FUTURES_REST_LONG_SHORT_RATIO_ENDPOINT + f"{instrument}"
        description = "FUTURES Long Short Request"
        lg.info(f"Starting {description}")

        if parallel_execution:
            if start_date is None or end_date is None:
                raise ValueError("Start and end date must be provided for parallel execution!")
            else:
                return_df = RestService._process_parallel(start_date, end_date, batch_period, self._headers(), url,
                                                          params, description, self._get_max_threads())
                # Check if timestamp or exchangeTimestamp + exchangeTimestampNano is present and sort by it otherwise skip sorting
                if 'timestamp' in return_df.columns:
                    return_df.sort_values('timestamp', inplace=True)
                elif 'exchangeTimestamp' in return_df.columns and 'exchangeTimestampNano' in return_df.columns:
                    return_df.sort_values(['exchangeTimestamp', 'exchangeTimestampNano'], inplace=True)
        else:
            return_df = RestService.get_and_process_response_df(url, params, self._headers(), description)

        if not return_df.empty and 'timestamp' in return_df.columns:
            return_df['timestamp'] = pd.to_datetime(return_df["timestamp"], unit="ms", utc=True)
        lg.info(f"Finished {description}")
        return return_df

    def get_long_short_ratio(self, instrument: str, exchange: MarketDataVenue = None, start_date: datetime = None,
                             end_date: datetime = None,
                             time_format: TimeFormat = None, time_interval: TimeInterval = None,
                             sort_direction: SortDirection = None,
                             index_keys: List[str] = None) -> pd.DataFrame:
        if index_keys is None:
            index_keys = ['exchange']
        return self.get_long_short_ratio_raw(instrument, exchange, start_date, end_date, time_format, time_interval,
                                             sort_direction).set_index(index_keys)

    def get_ohlcv_information_raw(self, exchanges: List[MarketDataVenue] = None, include_inactive: bool = False,
                                  time_format: TimeFormat = None) -> pd.DataFrame:
        params = {}
        if exchanges is not None and len(exchanges) > 0:
            params['exchange'] = ",".join(exchange.value for exchange in exchanges)
        if include_inactive:
            params['includeInactive'] = str(include_inactive).lower()
        if time_format is not None:
            params['timeFormat'] = time_format.value

        url = AMBERDATA_FUTURES_REST_OHLCV_ENDPOINT + f"information"
        description = "FUTURES OHLCV Information Request"
        lg.info(f"Starting {description}")
        raw_data = RestService.get_and_process_response_dict(url, params, self._headers(), description)
        lg.info(f"Finished {description}")

        # Process the raw data into a more structured format
        processed_data = []
        for entry in raw_data['data']:
            row = {
                'exchange': entry['exchange'],
                'instrument': entry['instrument'],
                'startDate': entry['startDate'],
                'endDate': entry['endDate']
            }
            processed_data.append(row)

        # Convert processed data into DataFrame
        processed_df = pd.DataFrame(processed_data)
        return processed_df

    def get_ohlcv_information(self, exchanges: List[MarketDataVenue] = None, include_inactive: bool = None,
                              time_format: TimeFormat = None,
                              index_keys: List[str] = None) -> pd.DataFrame:
        if index_keys is None:
            index_keys = ['exchange', 'instrument']
        processed_df = self.get_ohlcv_information_raw(exchanges, include_inactive, time_format)
        return processed_df.set_index(index_keys)

    def get_ohlcv_raw(self, instrument: str, exchanges: List[MarketDataVenue], start_date: datetime = None,
                      end_date: datetime = None, time_interval: TimeInterval = None,
                      time_format: TimeFormat = None, sort_direction: SortDirection = None) -> pd.DataFrame:
        all_processed_data = []

        for exchange in exchanges:
            params = {'exchange': exchange.value}
            if start_date is not None:
                params['startDate'] = start_date.isoformat()
            if end_date is not None:
                params['endDate'] = end_date.isoformat()
            if time_interval is not None:
                params['timeInterval'] = time_interval.value
            if time_format is not None:
                params['timeFormat'] = time_format.value
            if sort_direction is not None:
                params['sortDirection'] = sort_direction.value

            url = AMBERDATA_FUTURES_REST_OHLCV_ENDPOINT + f"{instrument}"
            description = f"FUTURES OHLCV Request for {instrument} on {exchange.value}"
            lg.info(f"Starting {description} with URL: {url} and params: {params}")

            raw_data = RestService.get_and_process_response_dict(url, params, self._headers(), description)
            lg.info(f"Finished {description}")

            # Check if 'data' key exists
            if not raw_data:
                lg.error("Raw response is empty.")
                raise NoDataReturned("Raw response is empty.")
            if 'data' not in raw_data:
                lg.error(f"Key 'data' not found in the response. Raw response: {json.dumps(raw_data, indent=2)}")
                raise NoDataReturned(
                    f"Key 'data' not found in the response. Raw response: {json.dumps(raw_data, indent=2)}")

            # Process the raw data into a more structured format
            processed_data = []
            for ohlcv in raw_data['data']:
                processed_data.append({
                    'timestamp': ohlcv['exchangeTimestamp'],
                    'open': ohlcv['open'],
                    'high': ohlcv['high'],
                    'low': ohlcv['low'],
                    'close': ohlcv['close'],
                    'volume': ohlcv['volume'],
                    'exchange': exchange.value,
                    'instrument': ohlcv['instrument']
                })

            all_processed_data.extend(processed_data)

        # Convert processed data into DataFrame
        processed_df = pd.DataFrame(all_processed_data)
        return processed_df

    def get_ohlcv(self, instrument: str, exchanges: List[MarketDataVenue], start_date: datetime = None,
                  end_date: datetime = None, time_interval: TimeInterval = None,
                  time_format: TimeFormat = None, sort_direction: SortDirection = None,
                  index_keys: List[str] = None) -> pd.DataFrame:
        if index_keys is None:
            index_keys = ['exchange', 'timestamp']
        processed_df = self.get_ohlcv_raw(instrument, exchanges, start_date, end_date, time_interval,
                                          time_format, sort_direction)
        return processed_df.set_index(index_keys)

    def get_ohlcv_batch_raw(self, exchanges: List[MarketDataVenue], instruments: List[str], start_date: datetime = None,
                            end_date: datetime = None, time_interval: TimeInterval = None,
                            time_format: TimeFormat = None) -> pd.DataFrame:
        all_processed_data = []

        for exchange in exchanges:
            params = {'instrument': ",".join(instrument for instrument in instruments)}
            if start_date is not None:
                params['startDate'] = start_date.isoformat()
            if end_date is not None:
                params['endDate'] = end_date.isoformat()
            if time_interval is not None:
                params['timeInterval'] = time_interval.value
            if time_format is not None:
                params['timeFormat'] = time_format.value

            url = AMBERDATA_FUTURES_REST_BATCH_OHLCV_ENDPOINT + f"exchange/{exchange}/historical"
            instrumentStr = ",".join(instrument for instrument in instruments)
            description = f"FUTURES OHLCV Request for {exchange} for {instrumentStr}"
            lg.info(f"Starting {description} with URL: {url} and params: {params}")

            raw_data = RestService.get_and_process_response_dict(url, params, self._headers(), description)
            lg.info(f"Finished {description}")

            # Check if 'data' key exists
            if not raw_data:
                lg.error("Raw response is empty.")
                raise NoDataReturned("Raw response is empty.")
            if 'data' not in raw_data:
                lg.error(f"Key 'data' not found in the response. Raw response: {json.dumps(raw_data, indent=2)}")
                raise NoDataReturned(
                    f"Key 'data' not found in the response. Raw response: {json.dumps(raw_data, indent=2)}")

            # Process the raw data into a more structured format
            processed_data = []
            for ohlcv in raw_data['data']:
                processed_data.append({
                    'timestamp': ohlcv['timestamp'],
                    'open': ohlcv['open'],
                    'high': ohlcv['high'],
                    'low': ohlcv['low'],
                    'close': ohlcv['close'],
                    'volume': ohlcv['volume'],
                    'exchange': exchange.value,
                    'instrument': ohlcv['instrument']
                })

            all_processed_data.extend(processed_data)

        # Convert processed data into DataFrame
        processed_df = pd.DataFrame(all_processed_data)
        return processed_df

    def get_ohlcv_batch(self, exchanges: List[MarketDataVenue], instruments: List[str], start_date: datetime = None,
                        end_date: datetime = None, time_interval: TimeInterval = None, time_format: TimeFormat = None,
                        index_keys: List[str] = None) -> pd.DataFrame:
        if index_keys is None:
            index_keys = ['exchange']
        processed_df = self.get_ohlcv_batch_raw(exchanges, instruments, start_date, end_date, time_interval,
                                                time_format)
        return processed_df.set_index(index_keys)

    def get_open_interest_information_raw(self, exchanges: List[MarketDataVenue] = None, include_inactive: bool = None,
                                          time_format: TimeFormat = None) -> pd.DataFrame:
        params = {}
        if exchanges is not None and len(exchanges) > 0:
            params['exchange'] = ",".join(
                [exchange.value for exchange in exchanges])
        if include_inactive is not None:
            params['includeInactive'] = str(include_inactive).lower()
        if time_format is not None:
            params['timeFormat'] = time_format.value
        url = AMBERDATA_FUTURES_REST_OPEN_INTEREST_ENDPOINT + "information"
        description = "FUTURES Open Interest Information Request"
        lg.info(f"Starting {description}")
        return_df = RestService.get_and_process_response_df(url, params, self._headers(), description)
        lg.info(f"Finished {description}")
        return return_df

    def get_open_interest_information(self, exchanges: List[MarketDataVenue] = None, include_inactive: bool = None,
                                      time_format: TimeFormat = None, index_keys: List[str] = None) -> pd.DataFrame:
        if index_keys is None:
            index_keys = ['exchange']
        return self.get_open_interest_information_raw(exchanges, include_inactive, time_format).set_index(index_keys)

    def get_open_interest_raw(self, instrument: str, exchange: MarketDataVenue, start_date: datetime = None,
                              end_date: datetime = None, time_format: TimeFormat = None,
                              sort_direction: SortDirection = None,
                              batch_period: datetime.timedelta = BatchPeriod.HOUR_1.value,
                              parallel_execution: bool = False) -> pd.DataFrame:
        params = {'exchange': exchange.value}
        if start_date is not None:
            params['startDate'] = start_date.isoformat(timespec='milliseconds')
        if end_date is not None:
            params['endDate'] = end_date.isoformat(timespec='milliseconds')
        if time_format is not None:
            params['timeFormat'] = time_format.value
        if sort_direction is not None:
            params['sortDirection'] = sort_direction.value
        url = AMBERDATA_FUTURES_REST_OPEN_INTEREST_ENDPOINT + f"{instrument}"
        description = "FUTURES Open Interest Request"
        lg.info(f"Starting {description}")

        if parallel_execution:
            if start_date is None or end_date is None:
                raise ValueError("Start and end date must be provided for parallel execution!")
            else:
                return_df = RestService._process_parallel(start_date, end_date, batch_period, self._headers(), url,
                                                          params, description, self._get_max_threads())
                # Check if timestamp or exchangeTimestamp + exchangeTimestampNano is present and sort by it otherwise skip sorting
                if 'timestamp' in return_df.columns:
                    return_df.sort_values('timestamp', inplace=True)
                elif 'exchangeTimestamp' in return_df.columns and 'exchangeTimestampNano' in return_df.columns:
                    return_df.sort_values(['exchangeTimestamp', 'exchangeTimestampNano'], inplace=True)
        else:
            return_df = RestService.get_and_process_response_df(url, params, self._headers(),
                                                                description)
        lg.info(f"Finished {description}")
        return return_df

    def get_open_interest(self, instrument: str, exchange: MarketDataVenue, start_date: datetime = None,
                          end_date: datetime = None, time_format: TimeFormat = None,
                          sort_direction: SortDirection = None, index_keys: List[str] = None,
                          batch_period: datetime.timedelta = BatchPeriod.HOUR_1.value,
                          parallel_execution: bool = False) -> pd.DataFrame:
        if index_keys is None:
            index_keys = ['exchange']
        return self.get_open_interest_raw(instrument, exchange, start_date, end_date, time_format,
                                          sort_direction, batch_period, parallel_execution).set_index(index_keys)

    def get_open_interest_batch_raw(self, exchanges: List[MarketDataVenue], instruments: List[str],
                                    start_date: datetime = None,
                                    end_date: datetime = None, time_interval: TimeInterval = None,
                                    time_format: TimeFormat = None) -> pd.DataFrame:
        all_processed_data = []

        for exchange in exchanges:
            params = {'instrument': ",".join(instrument for instrument in instruments)}
            if start_date is not None:
                params['startDate'] = start_date.isoformat()
            if end_date is not None:
                params['endDate'] = end_date.isoformat()
            if time_interval is not None:
                params['timeInterval'] = time_interval.value
            if time_format is not None:
                params['timeFormat'] = time_format.value

            url = AMBERDATA_FUTURES_REST_BATCH_OPEN_INTEREST_ENDPOINT + f"exchange/{exchange}/historical"
            instrumentStr = ",".join(instrument for instrument in instruments)
            description = f"FUTURES Open Interest Batch Request for {exchange} for {instrumentStr}"
            lg.info(f"Starting {description} with URL: {url} and params: {params}")

            raw_data = RestService.get_and_process_response_dict(url, params, self._headers(), description)
            lg.info(f"Finished {description}")

            # Check if 'data' key exists
            if not raw_data:
                lg.error("Raw response is empty.")
                raise NoDataReturned("Raw response is empty.")
            if 'data' not in raw_data:
                lg.error(f"Key 'data' not found in the response. Raw response: {json.dumps(raw_data, indent=2)}")
                raise NoDataReturned(
                    f"Key 'data' not found in the response. Raw response: {json.dumps(raw_data, indent=2)}")

            # Process the raw data into a more structured format
            processed_data = []
            for ohlcv in raw_data['data']:
                processed_data.append({
                    'timestamp': ohlcv['timestamp'],
                    'type': ohlcv['type'],
                    'value': ohlcv['value'],
                    'exchange': exchange.value,
                    'instrument': ohlcv['instrument']
                })

            all_processed_data.extend(processed_data)

        # Convert processed data into DataFrame
        processed_df = pd.DataFrame(all_processed_data)
        return processed_df

    def get_open_interest_batch(self, exchanges: List[MarketDataVenue], instruments: List[str],
                                start_date: datetime = None,
                                end_date: datetime = None, time_interval: TimeInterval = None,
                                time_format: TimeFormat = None,
                                index_keys: List[str] = None) -> pd.DataFrame:
        if index_keys is None:
            index_keys = ['exchange']
        processed_df = self.get_open_interest_batch_raw(exchanges, instruments, start_date, end_date, time_interval,
                                                        time_format)
        return processed_df.set_index(index_keys)

    def get_order_book_information_raw(self, exchanges: List[MarketDataVenue] = None, include_inactive: bool = None,
                                       time_format: TimeFormat = None) -> pd.DataFrame:
        params = {}
        if exchanges is not None and len(exchanges) > 0:
            params['exchange'] = ",".join(exchange.value for exchange in exchanges)
        if include_inactive is not None:
            params['includeInactive'] = str(include_inactive).lower()
        if time_format is not None:
            params['timeFormat'] = time_format.value if isinstance(time_format, TimeFormat) else time_format

        url = AMBERDATA_FUTURES_REST_ORDER_BOOK_SNAPSHOTS_ENDPOINT + "information"
        description = "FUTURES Order Book Information Request"
        lg.info(f"Starting {description}")
        return_df = RestService.get_and_process_response_df(url, params, self._headers(), description)
        lg.info(f"Finished {description}")
        return return_df

    def get_order_book_information(self, exchanges: List[MarketDataVenue] = None, include_inactive: bool = None,
                                   time_format: TimeFormat = None, index_keys: List[str] = None) -> pd.DataFrame:
        if index_keys is None:
            index_keys = ['exchange', 'instrument']
        return self.get_order_book_information_raw(exchanges, include_inactive, time_format).set_index(index_keys)

    def get_order_book_snapshots_historical_raw(self, instrument: str, exchange: MarketDataVenue,
                                                start_date: datetime = None, end_date: datetime = None,
                                                timestamp: datetime = None, time_format: TimeFormat = None,
                                                sort_direction: SortDirection = None) -> pd.DataFrame:
        params = {}
        params['exchange'] = exchange.value
        if start_date is not None:
            params['startDate'] = start_date.isoformat()
        if end_date is not None:
            params['endDate'] = end_date.isoformat()
        if sort_direction is not None:
            params['sortDirection'] = sort_direction
        if timestamp is not None:
            params['timestamp'] = timestamp.isoformat()
        if time_format is not None:
            params['timeFormat'] = time_format.value

        url = AMBERDATA_FUTURES_REST_ORDER_BOOK_SNAPSHOTS_ENDPOINT + f"{instrument}"
        description = f"FUTURES Order Book Snapshots Historical Request for {instrument}"
        lg.info(f"Starting {description}")
        raw_data = RestService.get_and_process_response_dict(url, params, self._headers(), description)
        lg.info(f"Finished {description}")

        # Process the raw data into a more structured format
        processed_data = []
        if 'data' in raw_data:
            for entry in raw_data['data']:
                timestamp = entry.get('timestamp')
                instrument = entry.get('instrument')
                exchange = entry.get('exchange')
                if 'ask' in entry:
                    for ask in entry['ask']:
                        processed_data.append({
                            'timestamp': timestamp,
                            'instrument': instrument,
                            'exchange': exchange,
                            'side': 'ask',
                            'price': ask.get('price'),
                            'volume': ask.get('volume'),
                            'numOrders': ask.get('numOrders')
                        })
                if 'bid' in entry:
                    for bid in entry['bid']:
                        processed_data.append({
                            'timestamp': timestamp,
                            'instrument': instrument,
                            'exchange': exchange,
                            'side': 'bid',
                            'price': bid.get('price'),
                            'volume': bid.get('volume'),
                            'numOrders': bid.get('numOrders')
                        })

        # Convert processed data into DataFrame
        processed_df = pd.DataFrame(processed_data)
        return processed_df

    def get_order_book_snapshots_historical(self, instrument: str, exchange: MarketDataVenue,
                                            start_date: datetime = None,
                                            end_date: datetime = None, timestamp: datetime = None,
                                            time_format: TimeFormat = None, sort_direction: SortDirection = None,
                                            index_keys: List[str] = None) -> pd.DataFrame:
        if index_keys is None:
            index_keys = ['timestamp', 'instrument', 'exchange', 'side']
        processed_df = self.get_order_book_snapshots_historical_raw(instrument, exchange, start_date, end_date,
                                                                    timestamp, time_format, sort_direction)
        return processed_df.set_index(index_keys)

    def get_order_book_events_historical_raw(self, instrument: str, exchange: MarketDataVenue,
                                             start_date: datetime = None,
                                             end_date: datetime = None, time_format: TimeFormat = None,
                                             sort_direction: SortDirection = None) -> pd.DataFrame:
        params = {'exchange': exchange.value}
        if start_date is not None:
            params['startDate'] = start_date.isoformat()
        if end_date is not None:
            params['endDate'] = end_date.isoformat()
        if time_format is not None:
            params['timeFormat'] = time_format.value
        if sort_direction is not None:
            params['sortDirection'] = sort_direction.value

        url = AMBERDATA_FUTURES_REST_ORDER_BOOK_EVENTS_ENDPOINT + f"{instrument}"
        description = f"FUTURES Order Book Events Historical Request for {instrument}"
        lg.info(f"Starting {description}")
        raw_data = RestService.get_and_process_response_dict(url, params, self._headers(), description)
        lg.info(f"Finished {description}")

        # Process the raw data into a more structured format
        processed_data = []
        if 'data' in raw_data:
            for entry in raw_data['data']:
                timestamp = entry.get('exchangeTimestamp')
                instrument = entry.get('instrument')
                exchange = entry.get('exchange')
                asks = entry.get('ask', [])
                bids = entry.get('bid', [])

                for ask in asks:
                    processed_data.append({
                        'timestamp': timestamp,
                        'instrument': instrument,
                        'exchange': exchange,
                        'side': 'ask',
                        'price': ask.get('price'),
                        'volume': ask.get('volume'),
                        'numOrders': ask.get('numOrders')
                    })
                for bid in bids:
                    processed_data.append({
                        'timestamp': timestamp,
                        'instrument': instrument,
                        'exchange': exchange,
                        'side': 'bid',
                        'price': bid.get('price'),
                        'volume': bid.get('volume'),
                        'numOrders': bid.get('numOrders')
                    })

        # Convert processed data into DataFrame
        processed_df = pd.DataFrame(processed_data)
        return processed_df

    def get_order_book_events_historical(self, instrument: str, exchange: MarketDataVenue, start_date: datetime = None,
                                         end_date: datetime = None, time_format: TimeFormat = None,
                                         sort_direction: SortDirection = None,
                                         index_keys: List[str] = None) -> pd.DataFrame:
        if index_keys is None:
            index_keys = ['timestamp', 'instrument', 'exchange', 'side']
        processed_df = self.get_order_book_events_historical_raw(instrument, exchange, start_date, end_date,
                                                                 time_format, sort_direction)
        return processed_df.set_index(index_keys)

    def get_tickers_information_raw(self, exchanges: List[MarketDataVenue] = None, include_inactive: bool = None,
                                    time_format: TimeFormat = None) -> pd.DataFrame:
        params = {}
        if exchanges is not None and len(exchanges) > 0:
            params['exchange'] = ",".join(exchange.value for exchange in exchanges)
        if include_inactive is not None:
            params['includeInactive'] = str(include_inactive).lower()
        if time_format is not None:
            params['timeFormat'] = time_format.value if isinstance(time_format, TimeFormat) else time_format

        url = AMBERDATA_FUTURES_REST_TICKERS_ENDPOINT + "information"
        description = "FUTURES Ticker Information Request"
        lg.info(f"Starting {description}")
        raw_response = RestService.get_and_process_response_dict(url, params, self._headers(), description)
        lg.info(f"Finished {description}")

        # Process raw data into DataFrame
        processed_data = []
        if 'data' in raw_response:
            for item in raw_response['data']:
                processed_data.append(item)

        return pd.DataFrame(processed_data)

    def get_tickers_information(self, exchanges: List[MarketDataVenue] = None, include_inactive: bool = None,
                                time_format: TimeFormat = None, index_keys: List[str] = None) -> pd.DataFrame:
        if index_keys is None:
            index_keys = ['exchange']
        return self.get_tickers_information_raw(exchanges, include_inactive, time_format).set_index(index_keys)

    def get_tickers_raw(self, instrument: str, exchange: MarketDataVenue,
                        start_date: datetime = None, end_date: datetime = None,
                        time_format: TimeFormat = None, sort_direction: SortDirection = None) -> pd.DataFrame:
        params = {
            'exchange': exchange.value,
        }
        if start_date is not None:
            params['startDate'] = start_date.isoformat(timespec='milliseconds')
        if end_date is not None:
            params['endDate'] = end_date.isoformat(timespec='milliseconds')
        if time_format is not None:
            params['timeFormat'] = time_format.value
        if sort_direction is not None:
            params['sortDirection'] = sort_direction.value
        url = AMBERDATA_FUTURES_REST_TICKERS_ENDPOINT + f"{instrument}"
        description = f"FUTURES Historical Ticker Request for {instrument}"
        lg.info(f"Starting {description}")
        return_df = RestService.get_and_process_response_df(url, params, self._headers(), description)
        lg.info(f"Finished {description}")
        return return_df

    def get_tickers(self, instrument: str, exchange: MarketDataVenue = None, start_date: datetime = None,
                    end_date: datetime = None, time_format: TimeFormat = None, sort_direction: SortDirection = None,
                    index_keys: List[str] = None) -> pd.DataFrame:
        if index_keys is None:
            index_keys = ['exchangeTimestamp', 'instrument', 'exchange']
        return self.get_tickers_raw(instrument, exchange, start_date, end_date, time_format, sort_direction).set_index(
            index_keys)

    def get_trades_information_raw(self, exchanges: List[MarketDataVenue] = None, include_inactive: bool = None,
                                   time_format: TimeFormat = None) -> pd.DataFrame:
        params = {}
        if include_inactive is not None:
            params['includeInactive'] = str(include_inactive).lower()
        if time_format is not None:
            params['timeFormat'] = time_format.value
        if exchanges is not None and len(exchanges) > 0:
            params['exchange'] = ",".join([exchange.value for exchange in exchanges])

        url = AMBERDATA_FUTURES_REST_TRADES_ENDPOINT + "information"
        description = "FUTURES Trades Information Request"
        lg.info(f"Starting {description}")
        return_df = RestService.get_and_process_response_df(url, params, self._headers(), description)
        lg.info(f"Finished {description}")
        return return_df

    def get_trades_information(self, exchanges: List[MarketDataVenue] = None, include_inactive: bool = None,
                               time_format: TimeFormat = None,
                               index_keys: List[str] = None) -> pd.DataFrame:
        if index_keys is None:
            index_keys = ['exchange', 'instrument']
        return self.get_trades_information_raw(exchanges, include_inactive, time_format).set_index(index_keys)

    def get_trades_raw(self, instrument: str, exchange: MarketDataVenue, start_date: datetime = None,
                       end_date: datetime = None,
                       time_format: TimeFormat = None, sort_direction: SortDirection = None) -> pd.DataFrame:
        params = {'exchange': exchange.value}
        if start_date is not None:
            params['startDate'] = start_date.isoformat()
        if end_date is not None:
            params['endDate'] = end_date.isoformat()
        if time_format is not None:
            params['timeFormat'] = time_format.value if isinstance(time_format, TimeFormat) else time_format
        if sort_direction is not None:
            params['sortDirection'] = sort_direction.value

        url = AMBERDATA_FUTURES_REST_TRADES_ENDPOINT + f"{instrument}"
        description = "FUTURES Trades Historical Request"
        lg.info(f"Starting {description}")
        raw_data = RestService.get_and_process_response_dict(url, params, self._headers(), description)
        lg.info(f"Finished {description}")

        # Process raw data into DataFrame
        processed_data = []
        if 'data' in raw_data:
            for entry in raw_data['data']:
                processed_data.append({
                    'timestamp': entry.get('exchangeTimestamp'),
                    'timestampNano': entry.get('exchangeTimestampNanoseconds'),
                    'instrument': instrument,
                    'side': "BUY" if entry.get('isBuySide') else "SELL",
                    'price': entry.get('price'),
                    'volume': entry.get('volume'),
                    'tradeId': entry.get('tradeId'),
                    'numOrders': entry.get('numOrders', 'unknown'),
                    'exchange': entry.get('exchange', 'unknown')  # Default to 'unknown' if exchange is missing
                })

        # Convert processed data into DataFrame
        processed_df = pd.DataFrame.from_records(processed_data)
        return processed_df

    def get_trades(self, instrument: str, exchange: MarketDataVenue, start_date: datetime = None,
                   end_date: datetime = None,
                   time_format: TimeFormat = None, sort_direction: SortDirection = None,
                   index_keys: List[str] = None) -> pd.DataFrame:
        if index_keys is None:
            index_keys = ['timestamp', 'instrument', 'exchange']
        return self.get_trades_raw(instrument, exchange, start_date, end_date, time_format, sort_direction).set_index(
            index_keys)



from datetime import timedelta
from enum import Enum

# REST API Endpoints
# Spot End Points
AMBERDATA_SPOT_REST_OHLCV_ENDPOINT = "https://api.amberdata.com/markets/spot/ohlcv/"
AMBERDATA_SPOT_REST_LEGACY_OHLCV_ENDPOINT = "https://api.amberdata.com/market/spot/ohlcv/"
AMBERDATA_SPOT_REST_TRADES_ENDPOINT = "https://api.amberdata.com/markets/spot/trades/"
AMBERDATA_SPOT_REST_PRICES_ENDPOINT = "https://api.amberdata.com/market/spot/prices/"
AMBERDATA_SPOT_REST_EXCHANGES_ENDPOINT = "https://api.amberdata.com/markets/spot/exchanges/information/"
AMBERDATA_SPOT_REST_PAIRS_ENDPOINT = "https://api.amberdata.com/market/spot/prices/pairs/information/"
AMBERDATA_SPOT_REST_EXCHANGES_REFERENCE_ENDPOINT = "https://api.amberdata.com/markets/spot/exchanges/reference/"
AMBERDATA_SPOT_REST_REFERENCE_RATES_ENDPOINT = "https://api.amberdata.com/markets/spot/reference-rates/"
AMBERDATA_SPOT_REST_TICKERS_ENDPOINT = "https://api.amberdata.com/markets/spot/tickers/"
AMBERDATA_SPOT_REST_TWAP_ENDPOINT = "https://api.amberdata.com/market/spot/twap/"
AMBERDATA_SPOT_REST_ORDER_BOOK_EVENTS_ENDPOINT = "https://api.amberdata.com/markets/spot/order-book-events/"
AMBERDATA_SPOT_REST_ORDER_BOOK_SNAPSHOTS_ENDPOINT = "https://api.amberdata.com/markets/spot/order-book-snapshots/"
AMBERDATA_SPOT_REST_VWAP_ENDPOINT = "https://api.amberdata.com/market/spot/vwap/"

# Futures End Points
AMBERDATA_FUTURES_REST_EXCHANGES_ENDPOINT = "https://api.amberdata.com//markets/futures/exchanges/"
AMBERDATA_FUTURES_REST_FUNDING_RATES_ENDPOINT = "https://api.amberdata.com/markets/futures/funding-rates/"
AMBERDATA_FUTURES_REST_BATCH_FUNDING_RATES_ENDPOINT = "https://api.amberdata.com/market/futures/funding-rates/"
AMBERDATA_FUTURES_REST_INSURANCE_FUNDS_ENDPOINT = "https://api.amberdata.com/markets/futures/insurance-fund/"
AMBERDATA_FUTURES_REST_LIQUIDATIONS_ENDPOINT = "https://api.amberdata.com/markets/futures/liquidations/"
AMBERDATA_FUTURES_REST_LONG_SHORT_RATIO_ENDPOINT = "https://api.amberdata.com/markets/futures/long-short-ratio/"
AMBERDATA_FUTURES_REST_OHLCV_ENDPOINT = "https://api.amberdata.com/markets/futures/ohlcv/"
AMBERDATA_FUTURES_REST_BATCH_OHLCV_ENDPOINT = "https://api.amberdata.com/market/futures/ohlcv/"
AMBERDATA_FUTURES_REST_OPEN_INTEREST_ENDPOINT = "https://api.amberdata.com/markets/futures/open-interest/"
AMBERDATA_FUTURES_REST_BATCH_OPEN_INTEREST_ENDPOINT = "https://api.amberdata.com/market/futures/open-interest/"
AMBERDATA_FUTURES_REST_ORDER_BOOK_SNAPSHOTS_ENDPOINT = "https://api.amberdata.com/markets/futures/order-book-snapshots/"
AMBERDATA_FUTURES_REST_ORDER_BOOK_EVENTS_ENDPOINT = "https://api.amberdata.com/markets/futures/order-book-events/"
AMBERDATA_FUTURES_REST_TICKERS_ENDPOINT = "https://api.amberdata.com/markets/futures/tickers/"
AMBERDATA_FUTURES_REST_TRADES_ENDPOINT = "https://api.amberdata.com/markets/futures/trades/"

# Swaps End Points
AMBERDATA_SWAPS_REST_BATCH_FUNDING_RATES_ENDPOINT = "https://api.amberdata.com/market/swaps/funding-rates/"
AMBERDATA_SWAPS_REST_BATCH_OHLCV_ENDPOINT = "https://api.amberdata.com/market/swaps/ohlcv/"
AMBERDATA_SWAPS_REST_BATCH_OPEN_INTEREST_ENDPOINT = "https://api.amberdata.com/market/swaps/open-interest/"

# Defi End Points
AMBERDATA_DEFI_REST_INFORMATION_ENDPOINT = "https://web3api.io/api/v2/market/defi/"
AMBERDATA_DEFI_REST_LENDING_ENDPOINT = "https://web3api.io/api/v2/market/defi/lending/exchanges/"
AMBERDATA_DEFI_REST_LIQUIDITY_ENDPOINT = "https://web3api.io/api/v2/market/defi/liquidity/"
AMBERDATA_DEFI_REST_LIQUIDITY_POOLS_ENDPOINT = "https://web3api.io/api/v2/market/defi/liquidity-positions/"
AMBERDATA_DEFI_REST_OHLCV_ENDPOINT = "https://web3api.io/api/v2/market/defi/ohlcv"
AMBERDATA_DEFI_REST_METRICS_ENDPOINT = "https://web3api.io/api/v2/market/defi/metrics/"
AMBERDATA_DEFI_REST_TRADES_ENDPOINT = "https://web3api.io/api/v2/market/defi/trades/"
AMBERDATA_DEFI_REST_TWAP_ENDPOINT = "https://web3api.io/api/v2/market/defi/twap/"
AMBERDATA_DEFI_REST_VWAP_ENDPOINT = "https://web3api.io/api/v2/market/defi/vwap/"
AMBERDATA_DEFI_REST_PRICES_ENDPOINT = "https://web3api.io/api/v2/market/defi/prices/"


class MarketDataVenue(str, Enum):
    BINANCE = "binance"
    COINBASE = "gdax"
    BITFINEX = "bitfinex"
    BITSTAMP = "bitstamp"
    GDAX = "gdax"
    BYBIT = "bybit"
    HUOBI = "huobi"
    BITMEX = "bitmex"
    OKEX = "okex"
    CBOE = "cboedigital"
    DERIBIT = "deribit"
    KRAKEN = "kraken"

class TimeFormat(Enum):
    MILLISECONDS = "milliseconds"
    MS = "ms"
    ISO = "iso"
    ISO8601 = "iso8601"
    HR = "hr"
    HUMAN_READABLE = "human_readable"

class DexDataVenue(str, Enum):
    UNISWAP_V2 = "uniswapv2"
    UNISWAP_V3 = "uniswapv3"
    SUSHISWAP = "sushiswap"
    BALANCER_VAULT = "balancer vault"
    CURVE_V1 = "curvev1"
    PANCAKESWAP = "Pancake LPs"
    CRODEFISWAP = "CroDefiSwap"

class LendingProtocol(str, Enum):
    AAVE_V1 = "aave_v1"
    AAVE_V2 = "aave_v2"
    AAVE_V3 = "aave_v3"
    COMPOUND_V1 = "compound_v1"
    CREAM_V1 = "cream_v1"
    CREAM_V2 = "cream_v2"
    FORTUBE_V1 = "fortube_v1"


class TimeInterval(Enum):
    MINUTE = 'minutes'
    HOUR = 'hours'
    DAY = 'days'
    TICKS = 'ticks'


class BatchPeriod(Enum):
    HOUR_1 = timedelta(hours=1)
    HOUR_2 = timedelta(hours=2)
    HOUR_4 = timedelta(hours=4)
    HOUR_8 = timedelta(hours=8)
    HOUR_12 = timedelta(hours=12)
    HOUR_16 = timedelta(hours=16)
    HOUR_20 = timedelta(hours=20)
    DAY_1 = timedelta(days=1)
    DAY_3 = timedelta(days=3)
    DAY_7 = timedelta(days=7)

class TimeBucket(Enum):
    MINUTES_5 = '5m'
    HOURS_1 = '1h'
    DAYS_1 = '1d'

class SortBy(Enum):
    NAME = 'name'
    NUMPAIRS = 'numPairs'

class SortDirection(Enum):
    ASCENDING = 'asc'
    DESCENDING = 'desc'

class DailyTime(Enum):
    T1600_M0500 = "T16:00-05:00"
    T1600_M0400 = "T16:00-04:00"
    T1600_P0000 = "T16:00+00:00"
    T1600_P0100 = "T16:00+01:00"
    T1600_P0400 = "T16:00+04:00"
    T1600_P0800 = "T16:00+08:00"
    T1600_P0900 = "T16:00+09:00"
